{% extends "!footer.html" %}
{%- block extrafooter %}
<button onclick="topFunction()" id="btn" title="Go to top">Top</button>

<script>
var mybutton = document.getElementById("btn");

window.onscroll = function() {scrollFunction()};

function scrollFunction() {
  if (document.body.scrollTop > 20 || document.documentElement.scrollTop > 20) {
    mybutton.style.display = "block";
  } else {
    mybutton.style.display = "none";
  }
}

function topFunction() {
  document.body.scrollTop = 0;
  document.documentElement.scrollTop = 0;
}
</script>
{% endblock %}
{#https://sidecar.gitter.im/#}

