@import url("css/theme.css");

img {
  border: none;
  display: block;
  margin-left: auto;
  margin-right: auto;
  text-align: center;
}
.wy-nav-content {
    max-width: none;
}

#btn {
  display: none; 
  position: fixed;
  bottom: 20px; 
  right: 30px; 
  z-index: 99;
  border: none; 
  outline: none;
  background-color: #848484;
  color: white; 
  cursor: pointer;
  padding: 15px; 
  border-radius: 10px; 
  font-size: 18px; 
}

